from typing import Literal, Optional

from pydantic import BaseModel, ConfigDict, Field

from ..types.common import EthAddress, HexString, Keccak256OrPadded


class TransactionLog(BaseModel):
    """Log entry in a transaction receipt."""

    address: EthAddress
    topics: list[Keccak256OrPadded]
    data: HexString
    block_number: int = Field(alias="blockNumber")
    transaction_hash: HexString = Field(alias="transactionHash")
    transaction_index: int = Field(alias="transactionIndex")
    block_hash: HexString = Field(alias="blockHash")
    block_timestamp: Optional[HexString] = Field(None, alias="blockTimestamp")
    log_index: int = Field(alias="logIndex")
    removed: bool


class TransactionReceipt(BaseModel):
    """Transaction receipt from the blockchain."""

    model_config = ConfigDict(populate_by_name=True)

    tx_hash: HexString = Field(alias="transactionHash")
    tx_index: int = Field(alias="transactionIndex")
    block_hash: HexString = Field(alias="blockHash")
    block_number: int = Field(alias="blockNumber")

    status: Literal[0, 1]
    type: int

    gas_used: int = Field(alias="gasUsed")
    cumulative_gas_used: int = Field(alias="cumulativeGasUsed")
    effective_gas_price: int = Field(alias="effectiveGasPrice")

    from_address: EthAddress = Field(alias="from")
    to_address: EthAddress = Field(alias="to")
    contract_address: Optional[EthAddress] = Field(None, alias="contractAddress")

    logs: list[TransactionLog]
    logs_bloom: HexString = Field(alias="logsBloom")



from __future__ import annotations

from collections.abc import Iterator
from datetime import UTC, datetime, timedelta
from typing import Any, cast

import httpx
import pytest

from polymarket_apis.testing.contract_assertions import (
    assert_api_contract,
    fail_contract,
    fetch_json,
)
from polymarket_apis.types.clob_types import ClobMarket, OrderBookSummary
from polymarket_apis.types.common import TimeseriesPoint
from polymarket_apis.types.data_types import (
    Activity,
    ClosedPosition,
    EventLiveVolume,
    HolderResponse,
    LeaderboardUser,
    MarketValue,
    Position,
    Trade,
    UserMetric,
    UserRank,
    ValueResponse,
)
from polymarket_apis.types.gamma_types import Event, GammaMarket


@pytest.fixture(scope="module")
def http_client() -> Iterator[httpx.Client]:
    with httpx.Client(http2=True, timeout=30.0) as client:
        yield client


@pytest.fixture(scope="module")
def gamma_markets_payload(http_client: httpx.Client) -> list[dict[str, Any]]:
    payload = fetch_json(
        "gamma /markets",
        http_client,
        "https://gamma-api.polymarket.com/markets",
        params={
            "limit": 5,
            "offset": 0,
            "active": True,
            "closed": False,
            "start_date_min": int(
                (datetime.now(tz=UTC) - timedelta(days=1)).timestamp()
            ),
        },
    )
    if not isinstance(payload, list):
        fail_contract("schema mismatch", "gamma /markets returned a non-list payload.")
    if not payload:
        fail_contract("endpoint unavailable", "gamma /markets returned an empty list.")
    return payload


@pytest.fixture(scope="module")
def gamma_market(gamma_markets_payload: list[dict[str, Any]]) -> GammaMarket:
    markets = cast(
        "list[GammaMarket]",
        assert_api_contract("gamma /markets", list[GammaMarket], gamma_markets_payload),
    )
    return markets[0]


@pytest.fixture(scope="module")
def gamma_events_payload(http_client: httpx.Client) -> list[dict[str, Any]]:
    payload = fetch_json(
        "gamma /events",
        http_client,
        "https://gamma-api.polymarket.com/events",
        params={
            "limit": 5,
            "offset": 0,
            "active": "true",
            "archived": "false",
            "start_date_min": int(
                (datetime.now(tz=UTC) - timedelta(days=1)).timestamp()
            ),
        },
    )
    if not isinstance(payload, list):
        fail_contract("schema mismatch", "gamma /events returned a non-list payload.")
    if not payload:
        fail_contract("endpoint unavailable", "gamma /events returned an empty list.")
    return payload


@pytest.fixture(scope="module")
def gamma_event(gamma_events_payload: list[dict[str, Any]]) -> Event:
    events = cast(
        "list[Event]",
        assert_api_contract("gamma /events", list[Event], gamma_events_payload),
    )
    return events[0]


@pytest.fixture(scope="module")
def leaderboard_payload(http_client: httpx.Client) -> list[dict[str, Any]]:
    payload = fetch_json(
        "data /v1/leaderboard",
        http_client,
        "https://data-api.polymarket.com/v1/leaderboard",
        params={
            "category": "OVERALL",
            "timePeriod": "DAY",
            "orderBy": "VOL",
            "limit": 5,
            "offset": 0,
        },
    )
    if not isinstance(payload, list):
        fail_contract(
            "schema mismatch",
            "data /v1/leaderboard returned a non-list payload.",
        )
    if not payload:
        fail_contract(
            "endpoint unavailable",
            "data /v1/leaderboard returned an empty list.",
        )
    return payload


@pytest.fixture(scope="module")
def leaderboard_user(leaderboard_payload: list[dict[str, Any]]) -> LeaderboardUser:
    users = cast(
        "list[LeaderboardUser]",
        assert_api_contract(
            "data /v1/leaderboard",
            list[LeaderboardUser],
            leaderboard_payload,
        ),
    )
    return users[0]


@pytest.mark.prod_read
def test_gamma_markets_schema(gamma_markets_payload: list[dict[str, Any]]) -> None:
    markets = assert_api_contract(
        "gamma /markets", list[GammaMarket], gamma_markets_payload
    )
    if not markets:
        fail_contract(
            "endpoint unavailable", "gamma /markets produced no validated markets."
        )


@pytest.mark.prod_read
def test_gamma_events_schema(gamma_events_payload: list[dict[str, Any]]) -> None:
    events = assert_api_contract("gamma /events", list[Event], gamma_events_payload)
    if not events:
        fail_contract(
            "endpoint unavailable", "gamma /events produced no validated events."
        )


@pytest.mark.prod_read
def test_clob_market_schema(
    http_client: httpx.Client, gamma_market: GammaMarket
) -> None:
    if not gamma_market.condition_id:
        fail_contract("schema mismatch", "Gamma market is missing condition_id.")
    payload = fetch_json(
        "clob /markets/{condition_id}",
        http_client,
        f"https://clob.polymarket.com/markets/{gamma_market.condition_id}",
    )
    market = assert_api_contract("clob /markets/{condition_id}", ClobMarket, payload)
    if market.condition_id != gamma_market.condition_id:
        fail_contract(
            "schema mismatch",
            "CLOB market condition_id did not match the source Gamma market condition_id.",
        )


@pytest.mark.prod_read
def test_clob_order_book_schema(
    http_client: httpx.Client,
    gamma_market: GammaMarket,
) -> None:
    if not gamma_market.token_ids:
        fail_contract("schema mismatch", "Gamma market is missing token ids.")
    token_ids = gamma_market.token_ids
    payload = fetch_json(
        "clob /book",
        http_client,
        "https://clob.polymarket.com/book",
        params={"token_id": token_ids[0]},
    )
    order_book = assert_api_contract("clob /book", OrderBookSummary, payload)
    if order_book.condition_id != gamma_market.condition_id:
        fail_contract(
            "schema mismatch",
            "CLOB order book condition_id did not match the source Gamma market condition_id.",
        )


@pytest.mark.prod_read
def test_data_positions_schema(
    http_client: httpx.Client,
    leaderboard_user: LeaderboardUser,
) -> None:
    payload = fetch_json(
        "data /positions",
        http_client,
        "https://data-api.polymarket.com/positions",
        params={
            "user": leaderboard_user.proxy_wallet,
            "sizeThreshold": 0,
            "limit": 5,
            "offset": 0,
        },
    )
    positions = assert_api_contract("data /positions", list[Position], payload)
    if not isinstance(positions, list):
        fail_contract("schema mismatch", "data /positions did not validate as a list.")


@pytest.mark.prod_read
def test_data_activity_schema(
    http_client: httpx.Client,
    leaderboard_user: LeaderboardUser,
) -> None:
    payload = fetch_json(
        "data /activity",
        http_client,
        "https://data-api.polymarket.com/activity",
        params={"user": leaderboard_user.proxy_wallet, "limit": 5, "offset": 0},
    )
    activity = assert_api_contract("data /activity", list[Activity], payload)
    if not isinstance(activity, list):
        fail_contract("schema mismatch", "data /activity did not validate as a list.")


@pytest.mark.prod_read
def test_data_trades_schema(
    http_client: httpx.Client,
    gamma_market: GammaMarket,
) -> None:
    if not gamma_market.condition_id:
        fail_contract("schema mismatch", "Gamma market is missing condition_id.")
    payload = fetch_json(
        "data /trades",
        http_client,
        "https://data-api.polymarket.com/trades",
        params={"market": gamma_market.condition_id, "limit": 5, "offset": 0},
    )
    trades = assert_api_contract("data /trades", list[Trade], payload)
    if not isinstance(trades, list):
        fail_contract("schema mismatch", "data /trades did not validate as a list.")


@pytest.mark.prod_read
def test_data_holders_schema(
    http_client: httpx.Client,
    gamma_market: GammaMarket,
) -> None:
    if not gamma_market.condition_id:
        fail_contract("schema mismatch", "Gamma market is missing condition_id.")
    payload = fetch_json(
        "data /holders",
        http_client,
        "https://data-api.polymarket.com/holders",
        params={"market": gamma_market.condition_id, "limit": 5, "min_balance": 1},
    )
    holders = assert_api_contract("data /holders", list[HolderResponse], payload)
    if not isinstance(holders, list):
        fail_contract("schema mismatch", "data /holders did not validate as a list.")


@pytest.mark.prod_read
def test_data_value_schema(
    http_client: httpx.Client,
    leaderboard_user: LeaderboardUser,
) -> None:
    payload = fetch_json(
        "data /value",
        http_client,
        "https://data-api.polymarket.com/value",
        params={"user": leaderboard_user.proxy_wallet},
    )
    values = assert_api_contract("data /value", list[ValueResponse], payload)
    if not values:
        fail_contract("endpoint unavailable", "data /value returned no rows.")


@pytest.mark.prod_read
def test_data_closed_positions_schema(
    http_client: httpx.Client,
    leaderboard_user: LeaderboardUser,
) -> None:
    payload = fetch_json(
        "data /closed-positions",
        http_client,
        "https://data-api.polymarket.com/closed-positions",
        params={"user": leaderboard_user.proxy_wallet},
    )
    positions = assert_api_contract(
        "data /closed-positions", list[ClosedPosition], payload
    )
    if not isinstance(positions, list):
        fail_contract(
            "schema mismatch",
            "data /closed-positions did not validate as a list.",
        )


@pytest.mark.prod_read
def test_data_traded_schema(
    http_client: httpx.Client,
    leaderboard_user: LeaderboardUser,
) -> None:
    payload = fetch_json(
        "data /traded",
        http_client,
        "https://data-api.polymarket.com/traded",
        params={"user": leaderboard_user.proxy_wallet},
    )
    if not isinstance(payload, dict):
        fail_contract("schema mismatch", "data /traded returned a non-object payload.")
    if "traded" not in payload:
        fail_contract(
            "schema mismatch", "data /traded payload is missing the `traded` field."
        )
    if not isinstance(payload["traded"], int):
        fail_contract("schema mismatch", "data /traded `traded` field is not an int.")


@pytest.mark.prod_read
def test_data_open_interest_schema(
    http_client: httpx.Client,
    gamma_market: GammaMarket,
) -> None:
    if not gamma_market.condition_id:
        fail_contract("schema mismatch", "Gamma market is missing condition_id.")
    payload = fetch_json(
        "data /oi",
        http_client,
        "https://data-api.polymarket.com/oi",
        params={"market": gamma_market.condition_id},
    )
    open_interest = assert_api_contract("data /oi", list[MarketValue], payload)
    if not isinstance(open_interest, list):
        fail_contract("schema mismatch", "data /oi did not validate as a list.")


@pytest.mark.prod_read
def test_data_live_volume_schema(
    http_client: httpx.Client,
    gamma_event: Event,
) -> None:
    if gamma_event.id is None:
        fail_contract("schema mismatch", "Gamma event is missing id.")
    payload = fetch_json(
        "data /live-volume",
        http_client,
        "https://data-api.polymarket.com/live-volume",
        params={"id": gamma_event.id},
    )
    live_volume = assert_api_contract(
        "data /live-volume", list[EventLiveVolume], payload
    )
    if not isinstance(live_volume, list):
        fail_contract(
            "schema mismatch", "data /live-volume did not validate as a list."
        )


@pytest.mark.prod_read
def test_user_pnl_schema(
    http_client: httpx.Client,
    leaderboard_user: LeaderboardUser,
) -> None:
    payload = fetch_json(
        "user-pnl /user-pnl",
        http_client,
        "https://user-pnl-api.polymarket.com/user-pnl",
        params={
            "user_address": leaderboard_user.proxy_wallet,
            "interval": "1d",
            "fidelity": "1h",
        },
    )
    points = assert_api_contract("user-pnl /user-pnl", list[TimeseriesPoint], payload)
    if not isinstance(points, list):
        fail_contract(
            "schema mismatch", "user-pnl /user-pnl did not validate as a list."
        )


@pytest.mark.prod_read
def test_lb_profit_schema(
    http_client: httpx.Client,
    leaderboard_user: LeaderboardUser,
) -> None:
    payload = fetch_json(
        "lb-api /profit",
        http_client,
        "https://lb-api.polymarket.com/profit",
        params={"address": leaderboard_user.proxy_wallet, "window": "1d", "limit": 1},
    )
    metrics = assert_api_contract("lb-api /profit", list[UserMetric], payload)
    if not metrics:
        fail_contract("endpoint unavailable", "lb-api /profit returned no rows.")


@pytest.mark.prod_read
def test_lb_rank_schema(
    http_client: httpx.Client,
    leaderboard_user: LeaderboardUser,
) -> None:
    payload = fetch_json(
        "lb-api /rank",
        http_client,
        "https://lb-api.polymarket.com/rank",
        params={
            "address": leaderboard_user.proxy_wallet,
            "window": "1d",
            "rankType": "vol",
        },
    )
    ranks = assert_api_contract("lb-api /rank", list[UserRank], payload)
    if not ranks:
        fail_contract("endpoint unavailable", "lb-api /rank returned no rows.")



import logging
from collections.abc import Callable, Mapping
from json import JSONDecodeError
from typing import Any, cast

from lomond import WebSocket
from lomond.events import Text
from lomond.persist import persist
from pydantic import BaseModel, ValidationError

from ..types.clob_types import ApiCreds
from ..types.websockets_types import (
    ActivityOrderMatchEvent,
    ActivityTradeEvent,
    AssetPriceSubscribeEvent,
    AssetPriceUpdateEvent,
    BestBidAskEvent,
    CommentEvent,
    LastTradePriceEvent,
    LiveDataEvents,
    MarketEvents,
    MarketResolvedEvent,
    NewMarketEvent,
    OrderBookSummaryEvent,
    OrderEvent,
    PriceChangeEvent,
    QuoteEvent,
    ReactionEvent,
    RequestEvent,
    SportsGameUpdate,
    TickSizeChangeEvent,
    TradeEvent,
    UserEvents,
)

logger = logging.getLogger(__name__)


MARKET_EVENT_CLASSES: Mapping[str, type[MarketEvents]] = {
    "book": OrderBookSummaryEvent,
    "price_change": PriceChangeEvent,
    "tick_size_change": TickSizeChangeEvent,
    "last_trade_price": LastTradePriceEvent,
    "best_bid_ask": BestBidAskEvent,
    "new_market": NewMarketEvent,
    "market_resolved": MarketResolvedEvent,
}

USER_EVENT_CLASSES: Mapping[str, type[UserEvents]] = {
    "order": OrderEvent,
    "trade": TradeEvent,
}

LIVE_DATA_EVENT_CLASSES: Mapping[str, type[LiveDataEvents]] = {
    "trades": ActivityTradeEvent,
    "orders_matched": ActivityOrderMatchEvent,
    "comment_created": CommentEvent,
    "comment_removed": CommentEvent,
    "reaction_created": ReactionEvent,
    "reaction_removed": ReactionEvent,
    "request_created": RequestEvent,
    "request_edited": RequestEvent,
    "request_canceled": RequestEvent,
    "request_expired": RequestEvent,
    "quote_created": QuoteEvent,
    "quote_edited": QuoteEvent,
    "quote_canceled": QuoteEvent,
    "quote_expired": QuoteEvent,
    "subscribe": AssetPriceSubscribeEvent,
    "update": AssetPriceUpdateEvent,
}

def parse_json(event: Text) -> object | None:
    if not event.text or event.text.isspace():
        return None
    try:
        return cast("object", event.json)
    except JSONDecodeError:
        logger.warning("Invalid json: %s", event.text)
        return None


def substitute_cls[T: BaseModel](cls: type[T], data: dict[str, Any]) -> T | None:
    try:
        return cls(**data)
    except ValidationError:
        logger.exception("Cannot initiate: %s with %s", cls, data)
        return None


def parse_event[T: BaseModel](
    message: object,
    classes: Mapping[str, type[T]],
    event_type_field: str,
) -> T | None:
    if message is None:
        return None
    if not isinstance(message, dict):
        logger.warning("Got %s instead of dict", message)
        return None

    typ_obj = message.get(event_type_field)
    typ = typ_obj if isinstance(typ_obj, str) else None
    if typ is None:
        logger.warning("Missing or invalid event type field '%s' in message: %s", event_type_field, message)
        return None

    cls = classes.get(typ)
    if cls is None:
        logger.warning("Unknown event type: %s", typ)
        return None

    return substitute_cls(cls, message)


def parse_market_event(text: Text) -> MarketEvents | list[OrderBookSummaryEvent] | None:
    message = parse_json(text)
    if isinstance(message, list):
        result: list[OrderBookSummaryEvent] = []
        for item in message:
            obj = substitute_cls(OrderBookSummaryEvent, item)
            if obj is not None:
                result.append(obj)
        return result

    return parse_event(message, MARKET_EVENT_CLASSES, "event_type")


def parse_user_event(text: Text) -> UserEvents | None:
    message = parse_json(text)

    return parse_event(message, USER_EVENT_CLASSES, "event_type")


def parse_live_data_event(text: Text) -> LiveDataEvents | None:
    message = parse_json(text)

    return parse_event(message, LIVE_DATA_EVENT_CLASSES, "type")

def parse_sports_event(text: Text) -> SportsGameUpdate | None:
    message = parse_json(text)

    if message is None:
        return None
    if not isinstance(message, dict):
        logger.warning("Got %s instead of dict", message)
        return None

    return substitute_cls(SportsGameUpdate, message)

def _default_process_market_event(text: Text) -> None:
    ev = parse_market_event(text)
    if ev is not None:
        print(ev, "\n")


def _default_process_user_event(text: Text) -> None:
    ev = parse_user_event(text)
    if ev is not None:
        print(ev, "\n")


def _default_process_live_data_event(text: Text) -> None:
    ev = parse_live_data_event(text)
    if ev is not None:
        print(ev, "\n")

def _default_process_sports_event(text: Text) -> None:
    ev = parse_sports_event(text)
    if ev is not None:
        print(ev, "\n")


class PolymarketWebsocketsClient:
    def __init__(self) -> None:
        self.url_market = "wss://ws-subscriptions-clob.polymarket.com/ws/market"
        self.url_user = "wss://ws-subscriptions-clob.polymarket.com/ws/user"
        self.url_live_data = "wss://ws-live-data.polymarket.com"

    def market_socket(
        self,
        token_ids: list[str],
        custom_feature_enabled: bool = True,
        process_event: Callable[[Text], None] = _default_process_market_event,
    ) -> None:
        """
        Connect to the market websocket and subscribe to market events for specific token IDs.

        Args:
            token_ids: List of token IDs to subscribe to
            custom_feature_enabled: Enables best_bid_ask, new_market and market_resolved event types
            process_event: Callback function to process received events

        """
        websocket = WebSocket(self.url_market)

        for event in persist(websocket):  # persist automatically reconnects
            if event.name == "ready":
                websocket.send_json(
                    assets_ids=token_ids, custom_feature_enabled=custom_feature_enabled
                )
            elif event.name == "text":
                process_event(cast("Text", event))

    def user_socket(
        self,
        creds: ApiCreds,
        process_event: Callable[[Text], None] = _default_process_user_event,
    ) -> None:
        """
        Connect to the user websocket and subscribe to user events.

        Args:
            creds: API credentials for authentication
            process_event: Callback function to process received events

        """
        websocket = WebSocket(self.url_user)

        for event in persist(websocket):
            if event.name == "ready":
                websocket.send_json(
                    auth=creds.model_dump(by_alias=True),
                )
            elif event.name == "text":
                process_event(cast("Text", event))

    def live_data_socket(
        self,
        subscriptions: list[dict[str, Any]],
        process_event: Callable[[Text], None] = _default_process_live_data_event,
    ) -> None:
        # info on how to subscribe found at https://github.com/Polymarket/real-time-data-client?tab=readme-ov-file#subscribe
        """
        Connect to the live data websocket and subscribe to specified events.

        Args:
            subscriptions: List of subscription configurations
            process_event: Callback function to process received events

        """
        websocket = WebSocket(self.url_live_data)

        for event in persist(websocket):
            if event.name == "ready":
                payload = {
                    "action": "subscribe",
                    "subscriptions": subscriptions,
                }

                websocket.send_json(**payload)

            elif event.name == "text":
                process_event(cast("Text", event))

    def sports_socket(self, process_event: Callable[[Text], None] = _default_process_sports_event) -> None:
        websocket = WebSocket("wss://sports-api.polymarket.com/ws")

        for event in persist(websocket):
            if event.name == "text":
                process_event(cast("Text", event))



import re
from collections.abc import Iterable
from typing import Any, Literal, NotRequired, TypedDict, cast

from ens.ens import ChecksumAddress
from eth_account import Account
from eth_account.datastructures import SignedMessage
from eth_account.messages import encode_defunct
from web3.constants import ADDRESS_ZERO
from web3.contract import Contract


def get_market_index(question_id: str) -> int:
    """Extract the market index from a question ID (last 2 hex characters)."""
    return int(question_id[-2:], 16)


def get_index_set(question_ids: list[str]) -> int:
    """Calculate bitwise index set from question IDs."""
    indices = [get_market_index(question_id) for question_id in question_ids]
    return sum(1 << index for index in set(indices))


INT_REGEX = re.compile(r"^u?int(\d*)$")
BYTES_REGEX = re.compile(r"^bytes(\d+)$")


def _pack_primitive(typ: str, val: Any) -> bytes:
    if isinstance(val, str) and val.startswith("0x"):
        raw = bytes.fromhex(val[2:])
    elif isinstance(val, bytes):
        raw = val
    else:
        raw = val

    if typ == "string":
        if not isinstance(raw, bytes | str):
            msg = "string value must be str or bytes"
            raise TypeError(msg)
        return raw.encode() if isinstance(raw, str) else raw
    if typ == "bytes":
        if isinstance(raw, int):
            msg = "bytes value must be hex/bytes/str"
            raise TypeError(msg)
        if isinstance(raw, str):
            return raw.encode()
        return raw

    m = BYTES_REGEX.match(typ)
    if m:
        n = int(m.group(1))
        if isinstance(raw, int):
            b = raw.to_bytes(n, "big")
        elif isinstance(raw, str) and raw.startswith("0x"):
            b = bytes.fromhex(raw[2:])
        elif isinstance(raw, bytes | bytearray):
            b = bytes(raw)
        else:
            msg = f"unsupported value for {typ}"
            raise TypeError(msg)
        if len(b) != n:
            if len(b) > n:
                return b[:n]
            return b.ljust(n, b"\x00")
        return b

    if typ == "address":
        if isinstance(raw, str) and raw.startswith("0x"):
            addr = raw[2:]
        elif isinstance(raw, str):
            addr = raw
        elif isinstance(raw, bytes | bytearray):
            return bytes(raw[-20:])
        else:
            msg = "address must be hex string or bytes"
            raise TypeError(msg)
        addr = addr.rjust(40, "0")[-40:]
        return bytes.fromhex(addr)

    m = INT_REGEX.match(typ)
    if m:
        bits = int(m.group(1)) if m.group(1) else 256
        size = bits // 8
        if isinstance(raw, bytes | bytearray):
            intval = int.from_bytes(raw, "big")
        elif isinstance(raw, str) and raw.startswith("0x"):
            intval = int(raw, 16)
        elif isinstance(raw, int):
            intval = raw
        else:
            msg = f"unsupported value for {typ}"
            raise TypeError(msg)
        if intval < 0:
            intval &= (1 << bits) - 1
        return intval.to_bytes(size, "big")

    msg = f"unsupported type {typ}"
    raise ValueError(msg)


class AbiPackedParam(TypedDict):
    type: str
    value: Any


def abi_encode_packed(*params: AbiPackedParam) -> bytes:
    """
    Takes in sequence of {'type': str, 'value': any}.

    returns: concatenated packed bytes (no 0x prefix).
    """
    parts: list[bytes] = []
    for p in params:
        typ = p["type"]
        val = p["value"]
        if typ.endswith("[]"):
            inner = typ[:-2]
            if not isinstance(val, Iterable):
                msg = f"expected array value for {typ}"
                raise TypeError(msg)
            for elem in val:
                parts.append(abi_encode_packed({"type": inner, "value": elem}))
            continue
        parts.append(_pack_primitive(typ, val))
    return b"".join(parts)


def split_signature(signature_hex: str) -> dict[str, Any]:
    """
    Split a signature into r, s, v components compatible with Safe factory.

    Args:
        signature_hex: Signature as hex string

    Returns:
        dict: Dictionary with r, s, v components properly formatted for the contract

    """
    # Remove 0x prefix if present
    signature_hex = signature_hex.removeprefix("0x")

    # Convert to bytes
    signature_bytes = bytes.fromhex(signature_hex)

    if len(signature_bytes) != 65:
        msg = "Invalid signature length"
        raise ValueError(msg)

    # Extract r, s, v
    r = signature_bytes[:32]
    s = signature_bytes[32:64]
    v = signature_bytes[64]

    if v < 27:
        if v in {0, 1}:
            v += 27
        else:
            msg = "Invalid signature v value"
            raise ValueError(msg)

    # Return properly formatted components for the contract:
    # - r and s as bytes32 (keep as bytes, web3.py will handle conversion)
    # - v as uint8 (integer)
    return {
        "r": r,  # bytes32
        "s": s,  # bytes32
        "v": v,  # uint8
    }


def create_safe_create_signature(
    account: Account, chain_id: Literal[137, 80002]
) -> str:
    """
    Create EIP-712 signature for Safe creation.

    Returns:
        str: The signature as hex string

    """
    # EIP-712 domain
    domain = {
        "name": "Polymarket Contract Proxy Factory",
        "chainId": chain_id,
        "verifyingContract": "0xaacFeEa03eb1561C4e67d661e40682Bd20E3541b",
    }

    # EIP-712 types
    types = {
        "CreateProxy": [
            {"name": "paymentToken", "type": "address"},
            {"name": "payment", "type": "uint256"},
            {"name": "paymentReceiver", "type": "address"},
        ]
    }

    # Values to sign
    values = {
        "paymentToken": ADDRESS_ZERO,
        "payment": 0,
        "paymentReceiver": ADDRESS_ZERO,
    }

    # Create the signature using eth_account
    signature: SignedMessage = account.sign_typed_data(domain, types, values)

    return signature.signature.hex()


class SafeTxn(TypedDict):
    to: ChecksumAddress
    value: int
    data: str
    operation: NotRequired[int]


def sign_safe_transaction(
    account: Account, safe: Contract, safe_txn: SafeTxn, nonce: int
) -> SignedMessage:
    safe_tx_gas = 0
    base_gas = 0
    gas_price = 0
    gas_token = ADDRESS_ZERO
    refund_receiver = ADDRESS_ZERO

    tx_hash_bytes = safe.functions.getTransactionHash(
        safe_txn["to"],
        safe_txn["value"],
        safe_txn["data"],
        safe_txn.get("operation", 0),
        safe_tx_gas,
        base_gas,
        gas_price,
        gas_token,
        refund_receiver,
        nonce,
    ).call()

    tx_hash_hex = tx_hash_bytes.hex()
    message = encode_defunct(hexstr=tx_hash_hex)
    return cast("SignedMessage", account.sign_message(message))


def get_packed_signature(signed: SignedMessage) -> bytes:
    r = signed.r
    s = signed.s
    v = signed.v

    match v:
        case 0 | 1:
            v += 31
        case 27 | 28:
            v += 4
        case _:
            msg = "Invalid signature V value"
            raise ValueError(msg)

    packed_sig = abi_encode_packed(
        {"type": "uint256", "value": r},
        {"type": "uint256", "value": s},
        {"type": "uint8", "value": v},
    )

    return packed_sig


def create_proxy_struct(
    from_address: str,
    to: str,
    data: str,
    tx_fee: str,
    gas_price: str,
    gas_limit: str,
    nonce: str,
    relay_hub_address: str,
    relay_address: str,
) -> bytes:
    """Create struct hash for proxy wallet signature."""
    relay_hub_prefix = b"rlx:"

    encoded_from = bytes.fromhex(from_address[2:])
    encoded_to = bytes.fromhex(to[2:])
    encoded_data = bytes.fromhex(data.removeprefix("0x"))
    encoded_tx_fee = int(tx_fee).to_bytes(32, byteorder="big")
    encoded_gas_price = int(gas_price).to_bytes(32, byteorder="big")
    encoded_gas_limit = int(gas_limit).to_bytes(32, byteorder="big")
    encoded_nonce = int(nonce).to_bytes(32, byteorder="big")
    encoded_relay_hub = bytes.fromhex(relay_hub_address[2:])
    encoded_relay = bytes.fromhex(relay_address[2:])

    struct = (
        relay_hub_prefix
        + encoded_from
        + encoded_to
        + encoded_data
        + encoded_tx_fee
        + encoded_gas_price
        + encoded_gas_limit
        + encoded_nonce
        + encoded_relay_hub
        + encoded_relay
    )

    return struct



"""
Polymarket APIs - Unified interface for Polymarket's various APIs.

This package provides a comprehensive interface to Polymarket's APIs including:
- CLOB (Central Limit Order Book) API for trading
- Gamma API for event and market information
- Data API for user information
- Web3 API for blockchain interactions
- WebSocket API for real-time data streams
- GraphQL API for flexible data queries
"""

from __future__ import annotations

from importlib import import_module
from typing import TYPE_CHECKING, Any

__version__ = "0.5.3"
__author__ = "Razvan Gheorghe"
__email__ = "razvan@gheorghe.me"

if TYPE_CHECKING:
    from .clients import (
        AsyncPolymarketGraphQLClient,
        PolymarketClobClient,
        PolymarketDataClient,
        PolymarketGammaClient,
        PolymarketGaslessWeb3Client,
        PolymarketGraphQLClient,
        PolymarketReadOnlyClobClient,
        PolymarketWeb3Client,
        PolymarketWebsocketsClient,
    )
    from .types.clob_types import ApiCreds, MarketOrderArgs, OrderArgs, OrderType

__all__ = [
    "ApiCreds",
    "AsyncPolymarketGraphQLClient",
    "MarketOrderArgs",
    "OrderArgs",
    "OrderType",
    "PolymarketClobClient",
    "PolymarketDataClient",
    "PolymarketGammaClient",
    "PolymarketGaslessWeb3Client",
    "PolymarketGraphQLClient",
    "PolymarketReadOnlyClobClient",
    "PolymarketWeb3Client",
    "PolymarketWebsocketsClient",
    "__author__",
    "__email__",
    "__version__",
]

_EXPORT_MAP = {
    "ApiCreds": ".types.clob_types",
    "AsyncPolymarketGraphQLClient": ".clients",
    "MarketOrderArgs": ".types.clob_types",
    "OrderArgs": ".types.clob_types",
    "OrderType": ".types.clob_types",
    "PolymarketClobClient": ".clients",
    "PolymarketDataClient": ".clients",
    "PolymarketGammaClient": ".clients",
    "PolymarketGaslessWeb3Client": ".clients",
    "PolymarketGraphQLClient": ".clients",
    "PolymarketReadOnlyClobClient": ".clients",
    "PolymarketWeb3Client": ".clients",
    "PolymarketWebsocketsClient": ".clients",
}


def __getattr__(name: str) -> Any:
    module_name = _EXPORT_MAP.get(name)
    if module_name is None:
        msg = f"module {__name__!r} has no attribute {name!r}"
        raise AttributeError(msg)

    module = import_module(module_name, __name__)
    value = getattr(module, name)
    globals()[name] = value
    return value



"""
Client modules for Polymarket APIs.

This module contains all the client classes for interacting with different
Polymarket APIs including CLOB, Data, Gamma, GraphQL, Web3, and WebSocket clients.
"""

from __future__ import annotations

from importlib import import_module
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .clob_client import PolymarketClobClient, PolymarketReadOnlyClobClient
    from .data_client import PolymarketDataClient
    from .gamma_client import PolymarketGammaClient
    from .graphql_client import (
        AsyncPolymarketGraphQLClient,
        PolymarketGraphQLClient,
    )
    from .web3_client import PolymarketGaslessWeb3Client, PolymarketWeb3Client
    from .websockets_client import (
        PolymarketWebsocketsClient,
        parse_live_data_event,
        parse_market_event,
        parse_sports_event,
        parse_user_event,
    )

__all__ = [
    "AsyncPolymarketGraphQLClient",
    "PolymarketClobClient",
    "PolymarketDataClient",
    "PolymarketGammaClient",
    "PolymarketGaslessWeb3Client",
    "PolymarketGraphQLClient",
    "PolymarketReadOnlyClobClient",
    "PolymarketWeb3Client",
    "PolymarketWebsocketsClient",
    "parse_live_data_event",
    "parse_market_event",
    "parse_sports_event",
    "parse_user_event",
]

_EXPORT_MAP = {
    "AsyncPolymarketGraphQLClient": ".graphql_client",
    "PolymarketClobClient": ".clob_client",
    "PolymarketDataClient": ".data_client",
    "PolymarketGammaClient": ".gamma_client",
    "PolymarketGaslessWeb3Client": ".web3_client",
    "PolymarketGraphQLClient": ".graphql_client",
    "PolymarketReadOnlyClobClient": ".clob_client",
    "PolymarketWeb3Client": ".web3_client",
    "PolymarketWebsocketsClient": ".websockets_client",
    "parse_live_data_event": ".websockets_client",
    "parse_market_event": ".websockets_client",
    "parse_sports_event": ".websockets_client",
    "parse_user_event": ".websockets_client",
}


def __getattr__(name: str) -> Any:
    module_name = _EXPORT_MAP.get(name)
    if module_name is None:
        msg = f"module {__name__!r} has no attribute {name!r}"
        raise AttributeError(msg)

    module = import_module(module_name, __name__)
    value = getattr(module, name)
    globals()[name] = value
    return value



from poly_eip712_structs import Address, EIP712Struct, String, Uint


class ClobAuth(EIP712Struct):  # type: ignore[misc]
    address = Address()
    timestamp = String()
    nonce = Uint()
    message = String()



import json
import random
import string
from datetime import datetime
from typing import Literal, Optional, Union
from urllib.parse import urljoin

import httpx

from ..types.common import EthAddress
from ..types.data_types import UserProfile
from ..types.gamma_types import (
    Comment,
    Event,
    GammaMarket,
    SearchResult,
    Series,
    Sport,
    Tag,
    TagRelation,
    Team,
)


def generate_random_id(length: int = 16) -> str:
    characters = string.ascii_letters + string.digits
    random_id = "".join(random.choices(characters, k=length))
    return random_id


class PolymarketGammaClient:
    def __init__(self, base_url: str = "https://gamma-api.polymarket.com"):
        self.base_url = base_url
        self.client = httpx.Client(http2=True, timeout=30.0)

    def _build_url(self, endpoint: str) -> str:
        return urljoin(self.base_url, endpoint)

    def search(
        self,
        query: str,
        cache: Optional[bool] = None,
        status: Optional[Literal["active", "resolved"]] = None,
        limit_per_type: Optional[int] = None,  # max is 50
        page: Optional[int] = None,
        tags: Optional[list[str]] = None,
        keep_closed_markets: Optional[bool] = None,
        sort: Optional[
            Literal[
                "volume",
                "volume_24hr",
                "liquidity",
                "start_date",
                "end_date",
                "competitive",
            ]
        ] = None,
        ascending: Optional[bool] = None,
        search_tags: Optional[bool] = None,
        search_profiles: Optional[bool] = None,
        recurrence: Optional[
            Literal["hourly", "daily", "weekly", "monthly", "annual"]
        ] = None,
        exclude_tag_ids: Optional[list[int]] = None,
        optimized: Optional[bool] = None,
    ) -> SearchResult:
        params: dict[str, str | list[str] | int | bool] = {
            "q": query,
        }
        if cache is not None:
            params["cache"] = str(cache).lower()
        if status:
            params["events_status"] = status
        if limit_per_type:
            params["limit_per_type"] = limit_per_type
        if page:
            params["page"] = page
        if tags:
            params["events_tag"] = json.dumps([json.dumps(item) for item in tags])
        if keep_closed_markets is not None:
            params["keep_closed_markets"] = keep_closed_markets
        if sort:
            params["sort"] = sort
        if ascending is not None:
            params["ascending"] = str(ascending).lower()
        if search_tags is not None:
            params["search_tags"] = str(search_tags).lower()
        if search_profiles is not None:
            params["search_profiles"] = str(search_profiles).lower()
        if recurrence:
            params["recurrence"] = recurrence
        if exclude_tag_ids:
            params["exclude_tag_id"] = [str(i) for i in exclude_tag_ids]
        if optimized is not None:
            params["optimized"] = str(optimized).lower()
        response = self.client.get(self._build_url("/public-search"), params=params)
        response.raise_for_status()
        return SearchResult(**response.json())

    def get_market(self, market_id: str) -> GammaMarket:
        """Get a GammaMarket by market_id."""
        response = self.client.get(self._build_url(f"/markets/{market_id}"))
        response.raise_for_status()
        return GammaMarket(**response.json())

    def get_markets(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order: str | None = None,
        ascending: bool = True,
        archived: bool | None = None,
        active: bool | None = None,
        closed: bool | None = None,
        slugs: list[str] | None = None,
        market_ids: list[int] | None = None,
        token_ids: list[str] | None = None,
        condition_ids: list[str] | None = None,
        tag_id: int | None = None,
        related_tags: bool | None = False,
        liquidity_num_min: float | None = None,
        liquidity_num_max: float | None = None,
        volume_num_min: float | None = None,
        volume_num_max: float | None = None,
        start_date_min: datetime | None = None,
        start_date_max: datetime | None = None,
        end_date_min: datetime | None = None,
        end_date_max: datetime | None = None,
    ) -> list[GammaMarket]:
        params: dict[str, float | int | list[int] | str | list[str] | bool] = {}
        if limit:
            params["limit"] = limit
        if offset:
            params["offset"] = offset
        if order:
            params["order"] = order
            params["ascending"] = ascending
        if slugs:
            params["slug"] = slugs
        if archived is not None:
            params["archived"] = archived
        if active is not None:
            params["active"] = active
        if closed is not None:
            params["closed"] = closed
        if market_ids:
            params["id"] = market_ids
        if token_ids:
            params["clob_token_ids"] = token_ids
        if condition_ids:
            params["condition_ids"] = condition_ids
        if liquidity_num_min:
            params["liquidity_num_min"] = liquidity_num_min
        if liquidity_num_max:
            params["liquidity_num_max"] = liquidity_num_max
        if volume_num_min:
            params["volume_num_min"] = volume_num_min
        if volume_num_max:
            params["volume_num_max"] = volume_num_max
        if start_date_min:
            params["start_date_min"] = int(start_date_min.timestamp())
        if start_date_max:
            params["start_date_max"] = int(start_date_max.timestamp())
        if end_date_min:
            params["end_date_min"] = int(end_date_min.timestamp())
        if end_date_max:
            params["end_date_max"] = int(end_date_max.timestamp())
        if tag_id:
            params["tag_id"] = tag_id
            if related_tags:
                params["related_tags"] = related_tags

        response = self.client.get(self._build_url("/markets"), params=params)
        response.raise_for_status()
        return [GammaMarket(**market) for market in response.json()]

    def get_market_by_id(
        self, market_id: str, include_tag: Optional[bool] = None
    ) -> GammaMarket:
        params = {}
        if include_tag:
            params["include_tag"] = include_tag
        response = self.client.get(
            self._build_url(f"/markets/{market_id}"), params=params
        )
        response.raise_for_status()
        return GammaMarket(**response.json())

    def get_market_tags(self, market_id: str) -> list[Tag]:
        response = self.client.get(self._build_url(f"/markets/{market_id}/tags"))
        response.raise_for_status()
        return [Tag(**tag) for tag in response.json()]

    def get_market_by_slug(
        self, slug: str, include_tag: Optional[bool] = None
    ) -> GammaMarket:
        params = {}
        if include_tag:
            params["include_tag"] = include_tag
        response = self.client.get(
            self._build_url(f"/markets/slug/{slug}"), params=params
        )
        response.raise_for_status()
        return GammaMarket(**response.json())

    def get_events(
        self,
        limit: int = 500,
        offset: int = 0,
        order: Optional[str] = None,
        ascending: bool = True,
        event_ids: Optional[Union[str, list[str]]] = None,
        slugs: Optional[list[str]] = None,
        archived: Optional[bool] = None,
        active: Optional[bool] = None,
        closed: Optional[bool] = None,
        liquidity_min: Optional[float] = None,
        liquidity_max: Optional[float] = None,
        volume_min: Optional[float] = None,
        volume_max: Optional[float] = None,
        start_date_min: Optional[datetime] = None,
        start_date_max: Optional[datetime] = None,
        end_date_min: Optional[datetime] = None,
        end_date_max: Optional[datetime] = None,
        tag: Optional[str] = None,
        tag_id: Optional[int] = None,
        tag_slug: Optional[str] = None,
        related_tags: bool = False,
    ) -> list[Event]:
        params: dict[str, int | str | list[str] | float] = {
            "limit": limit,
            "offset": offset,
        }
        if order:
            params["order"] = order
            params["ascending"] = ascending
        if event_ids:
            params["id"] = event_ids
        if slugs:
            params["slug"] = slugs
        if archived is not None:
            params["archived"] = archived
        if active is not None:
            params["active"] = active
        if closed is not None:
            params["closed"] = closed
        if liquidity_min:
            params["liquidity_min"] = liquidity_min
        if liquidity_max:
            params["liquidity_max"] = liquidity_max
        if volume_min:
            params["volume_min"] = volume_min
        if volume_max:
            params["volume_max"] = volume_max
        if start_date_min:
            params["start_date_min"] = int(start_date_min.timestamp())
        if start_date_max:
            params["start_date_max"] = int(start_date_max.timestamp())
        if end_date_min:
            params["end_date_min"] = int(end_date_min.timestamp())
        if end_date_max:
            params["end_date_max"] = int(end_date_max.timestamp())
        if tag:
            params["tag"] = tag
        elif tag_id:
            params["tag_id"] = tag_id
            if related_tags:
                params["related_tags"] = related_tags
        elif tag_slug:
            params["tag_slug"] = tag_slug

        response = self.client.get(self._build_url("/events"), params=params)
        response.raise_for_status()
        return [Event(**event) for event in response.json()]

    def get_all_events(
        self,
        order: Optional[str] = None,
        ascending: bool = True,
        event_ids: Optional[Union[str, list[str]]] = None,
        slugs: Optional[list[str]] = None,
        archived: Optional[bool] = None,
        active: Optional[bool] = None,
        closed: Optional[bool] = None,
        liquidity_min: Optional[float] = None,
        liquidity_max: Optional[float] = None,
        volume_min: Optional[float] = None,
        volume_max: Optional[float] = None,
        start_date_min: Optional[datetime] = None,
        start_date_max: Optional[datetime] = None,
        end_date_min: Optional[datetime] = None,
        end_date_max: Optional[datetime] = None,
        tag: Optional[str] = None,
        tag_id: Optional[int] = None,
        tag_slug: Optional[str] = None,
        related_tags: bool = False,
    ) -> list[Event]:
        offset = 0
        events = []

        while True:
            part = self.get_events(
                offset=offset,
                order=order,
                ascending=ascending,
                event_ids=event_ids,
                slugs=slugs,
                archived=archived,
                active=active,
                closed=closed,
                liquidity_min=liquidity_min,
                liquidity_max=liquidity_max,
                volume_min=volume_min,
                volume_max=volume_max,
                start_date_min=start_date_min,
                start_date_max=start_date_max,
                end_date_min=end_date_min,
                end_date_max=end_date_max,
                tag=tag,
                tag_id=tag_id,
                tag_slug=tag_slug,
                related_tags=related_tags,
            )
            events.extend(part)

            if len(part) < 500:
                break

            offset += 500

        return events

    def get_event_by_id(
        self,
        event_id: str | int,
        include_chat: Optional[bool] = None,
        include_template: Optional[bool] = None,
    ) -> Event:
        params = {}
        if include_chat:
            params["include_chat"] = include_chat
        if include_template:
            params["include_template"] = include_template
        response = self.client.get(
            self._build_url(f"/events/{event_id}"), params=params
        )
        response.raise_for_status()
        return Event(**response.json())

    def get_event_by_slug(
        self,
        slug: str,
        include_chat: Optional[bool] = None,
        include_template: Optional[bool] = None,
    ) -> Event:
        params = {}
        if include_chat:
            params["include_chat"] = include_chat
        if include_template:
            params["include_template"] = include_template
        response = self.client.get(
            self._build_url(f"/events/slug/{slug}"), params=params
        )
        response.raise_for_status()
        return Event(**response.json())

    def get_event_tags(self, event_id: str | int) -> list[Tag]:
        response = self.client.get(self._build_url(f"/events/{event_id}/tags"))
        response.raise_for_status()
        return [Tag(**tag) for tag in response.json()]

    def get_teams(
        self,
        limit: int = 500,
        offset: int = 0,
        order: Optional[
            Literal[
                "id",
                "name",
                "league",
                "record",
                "logo",
                "abbreviation",
                "alias",
                "createdAt",
                "updatedAt",
            ]
        ] = None,
        ascending: bool = True,
        league: Optional[str] = None,
        name: Optional[str] = None,
        abbreviation: Optional[str] = None,
    ) -> list[Team]:
        params: dict[str, int | str] = {
            "limit": limit,
            "offset": offset,
        }
        if order:
            params["order"] = order
            params["ascending"] = str(ascending).lower()
        if league:
            params["league"] = league.lower()
        if name:
            params["name"] = name
        if abbreviation:
            params["abbreviation"] = abbreviation.lower()
        response = self.client.get(self._build_url("/teams"), params=params)
        response.raise_for_status()
        return [Team(**team) for team in response.json()]

    def get_all_teams(
        self,
        order: Optional[
            Literal[
                "id",
                "name",
                "league",
                "record",
                "logo",
                "abbreviation",
                "alias",
                "createdAt",
                "updatedAt",
            ]
        ] = None,
        ascending: bool = True,
        league: Optional[str] = None,
        name: Optional[str] = None,
        abbreviation: Optional[str] = None,
    ) -> list[Team]:
        offset = 0
        teams = []

        while True:
            part = self.get_teams(
                offset=offset,
                order=order,
                ascending=ascending,
                league=league,
                name=name,
                abbreviation=abbreviation,
            )
            teams.extend(part)

            if len(part) < 500:
                break

            offset += 500

        return teams

    def get_sports_metadata(
        self,
    ) -> list[Sport]:
        response = self.client.get(self._build_url("/sports"))
        response.raise_for_status()
        return [Sport(**sport) for sport in response.json()]

    def get_tags(
        self,
        limit: int = 300,
        offset: int = 0,
        order: Optional[
            Literal[
                "id",
                "label",
                "slug",
                "forceShow",
                "forceHide",
                "isCarousel",
                "createdAt",
                "updatedAt",
                "createdBy",
                "updatedBy",
            ]
        ] = None,
        ascending: bool = True,
        include_templates: Optional[bool] = None,
        is_carousel: Optional[bool] = None,
    ) -> list[Tag]:
        params: dict[str, int | str] = {
            "limit": limit,
            "offset": offset,
        }
        if order:
            params["order"] = order
            params["ascending"] = str(ascending).lower()
        if include_templates is not None:
            params["include_templates"] = str(include_templates).lower()
        if is_carousel is not None:
            params["is_carousel"] = str(is_carousel).lower()
        response = self.client.get(self._build_url("/tags"), params=params)
        response.raise_for_status()
        return [Tag(**tag) for tag in response.json()]

    def get_all_tags(
        self,
        order: Optional[
            Literal[
                "id",
                "label",
                "slug",
                "forceShow",
                "forceHide",
                "isCarousel",
                "createdAt",
                "updatedAt",
                "createdBy",
                "updatedBy",
            ]
        ] = None,
        ascending: bool = True,
        include_templates: Optional[bool] = None,
        is_carousel: Optional[bool] = None,
    ) -> list[Tag]:
        offset = 0
        tags = []

        while True:
            part = self.get_tags(
                offset=offset,
                order=order,
                ascending=ascending,
                include_templates=include_templates,
                is_carousel=is_carousel,
            )
            tags.extend(part)

            if len(part) < 300:
                break

            offset += 300

        return tags

    def get_tag(self, tag_id: str, include_template: Optional[bool] = None) -> Tag:
        params = {}
        if include_template is not None:
            params = {"include_template": str(include_template).lower()}
        response = self.client.get(self._build_url(f"/tags/{tag_id}"), params=params)
        response.raise_for_status()
        return Tag(**response.json())

    def get_related_tag_ids_by_tag_id(
        self,
        tag_id: int,
        omit_empty: Optional[bool] = None,
        status: Optional[Literal["active", "closed", "all"]] = None,
    ) -> list[TagRelation]:
        params = {}
        if omit_empty is not None:
            params["omit_empty"] = str(omit_empty).lower()
        if status:
            params["status"] = status
        response = self.client.get(
            self._build_url(f"/tags/{tag_id}/related-tags"), params=params
        )
        response.raise_for_status()
        return [TagRelation(**tag) for tag in response.json()]

    def get_related_tag_ids_by_slug(
        self,
        slug: str,
        omit_empty: Optional[bool] = None,
        status: Optional[Literal["active", "closed", "all"]] = None,
    ) -> list[TagRelation]:
        params = {}
        if omit_empty is not None:
            params["omit_empty"] = str(omit_empty).lower()
        if status:
            params["status"] = status
        response = self.client.get(
            self._build_url(f"/tags/slug/{slug}/related-tags"), params=params
        )
        response.raise_for_status()
        return [TagRelation(**tag) for tag in response.json()]

    def get_related_tags_by_tag_id(
        self,
        tag_id: int,
        omit_empty: Optional[bool] = None,
        status: Optional[Literal["active", "closed", "all"]] = None,
    ) -> list[Tag]:
        params = {}
        if omit_empty is not None:
            params["omit_empty"] = str(omit_empty).lower()
        if status:
            params["status"] = status
        response = self.client.get(
            self._build_url(f"/tags/{tag_id}/related-tags/tags"), params=params
        )
        response.raise_for_status()
        return [Tag(**tag) for tag in response.json()]

    def get_related_tags_by_slug(
        self,
        slug: str,
        omit_empty: Optional[bool] = None,
        status: Optional[Literal["active", "closed", "all"]] = None,
    ) -> list[Tag]:
        params = {}
        if omit_empty is not None:
            params["omit_empty"] = str(omit_empty).lower()
        if status:
            params["status"] = status
        response = self.client.get(
            self._build_url(f"/tags/slug/{slug}/related-tags/tags"), params=params
        )
        response.raise_for_status()
        return [Tag(**tag) for tag in response.json()]

    def get_series(
        self,
        limit: int = 300,
        offset: int = 0,
        order: Optional[str] = None,
        ascending: bool = True,
        slug: Optional[str] = None,
        closed: Optional[bool] = None,
        include_chat: Optional[bool] = None,
        recurrence: Optional[
            Literal[
                "hourly", "daily", "weekly", "monthly", "annual"
            ]  # results also contain "15m" but the server returns a 422 Unprocessable Content
        ] = None,
    ) -> list[Series]:
        params: dict[str, str | int | list[int]] = {
            "limit": limit,
            "offset": offset,
        }
        if order:
            params["order"] = order
            params["ascending"] = str(ascending).lower()
        if slug:
            params["slug"] = slug
        if closed is not None:
            params["closed"] = str(closed).lower()
        if include_chat is not None:
            params["include_chat"] = str(include_chat).lower()
        if recurrence is not None:
            params["recurrence"] = str(recurrence).lower()

        response = self.client.get(self._build_url("/series"), params=params)
        response.raise_for_status()
        return [Series(**series) for series in response.json()]

    def get_all_series(
        self,
        order: Optional[str] = None,
        ascending: bool = True,
        slug: Optional[str] = None,
        closed: Optional[bool] = None,
        include_chat: Optional[bool] = None,
        recurrence: Optional[
            Literal["hourly", "daily", "weekly", "monthly", "annual"]
        ] = None,
    ) -> list[Series]:
        offset = 0
        series = []

        while True:
            part = self.get_series(
                offset=offset,
                order=order,
                ascending=ascending,
                slug=slug,
                closed=closed,
                include_chat=include_chat,
                recurrence=recurrence,
            )
            series.extend(part)

            if len(part) < 300:
                break

            offset += 300
        return series

    def get_series_by_id(self, series_id: str) -> Series:
        response = self.client.get(self._build_url(f"/series/{series_id}"))
        response.raise_for_status()
        return Series(**response.json())

    def get_comments(
        self,
        parent_entity_type: Literal["Event", "Series", "market"],
        parent_entity_id: int,
        limit: int = 500,
        offset: int = 0,
        order: Optional[str] = None,
        ascending: bool = True,
        get_positions: Optional[bool] = None,
        holders_only: Optional[bool] = None,
    ) -> list[Comment]:
        """Warning, the server doesn't give back the right amount of comments you asked for."""
        params: dict[str, str | int] = {
            "parent_entity_type": parent_entity_type,
            "parent_entity_id": parent_entity_id,
            "limit": limit,
            "offset": offset,
        }
        if order:
            params["order"] = order
            params["ascending"] = str(ascending).lower()
        if get_positions is not None:
            params["get_positions"] = str(get_positions).lower()
        if holders_only is not None:
            params["holders_only"] = str(holders_only).lower()
        response = self.client.get(self._build_url("/comments"), params=params)
        response.raise_for_status()
        return [Comment(**comment) for comment in response.json()]

    def get_comments_by_id(
        self, comment_id: str, get_positions: Optional[bool] = None
    ) -> list[Comment]:
        """Returns all comments that belong to the comment's thread."""
        params = {}
        if get_positions is not None:
            params["get_positions"] = str(get_positions).lower()
        response = self.client.get(
            self._build_url(f"/comments/{comment_id}"), params=params
        )
        response.raise_for_status()
        return [Comment(**comment) for comment in response.json()]

    def get_comments_by_user_address(
        self,
        user_address: EthAddress,  # warning, this is the base address, not the proxy address
        limit: int = 500,
        offset: int = 0,
        order: Optional[str] = None,
        ascending: bool = True,
    ) -> list[Comment]:
        params: dict[str, str | int] = {
            "limit": limit,
            "offset": offset,
        }
        if order:
            params["order"] = order
            params["ascending"] = str(ascending).lower()
        response = self.client.get(
            self._build_url(f"/comments/user_address/{user_address}"), params=params
        )
        response.raise_for_status()
        return [Comment(**comment) for comment in response.json()]

    def get_public_profile(self, address: EthAddress) -> UserProfile:
        params = {"address": address}
        response = self.client.get(self._build_url("/public-profile"), params=params)
        response.raise_for_status()
        return UserProfile(**response.json())

    def grok_event_summary(self, event_slug: str) -> None:
        json_payload = {
            "id": generate_random_id(),
            "messages": [{"role": "user", "content": "", "parts": []}],
        }

        params = {
            "prompt": event_slug,
        }

        with self.client.stream(
            method="POST",
            url="https://polymarket.com/api/grok/event-summary",
            params=params,
            json=json_payload,
        ) as stream:
            messages = []
            citations = []
            seen_urls = set()

            for line_bytes in stream.iter_lines():
                line = (
                    line_bytes.decode() if isinstance(line_bytes, bytes) else line_bytes
                )
                if line.startswith("__SOURCES__:"):
                    sources_json_str = line[len("__SOURCES__:") :]
                    try:
                        sources_obj = json.loads(sources_json_str)
                        for source in sources_obj.get("sources", []):
                            url = source.get("url")
                            if url and url not in seen_urls:
                                citations.append(source)
                                seen_urls.add(url)
                    except json.JSONDecodeError:
                        pass
                else:
                    messages.append(line)
                    print(line, end="")  # or handle message text as needed

        # After reading streamed lines:
        if citations:
            print("\n\nSources:")
            for source in citations:
                print(f"- {source.get('url', 'Unknown URL')}")

    def grok_election_market_explanation(
        self, candidate_name: str, election_title: str
    ) -> None:
        text = f"Provide candidate information for {candidate_name} in the {election_title} on Polymarket."
        json_payload = {
            "id": generate_random_id(),
            "messages": [
                {
                    "role": "user",
                    "content": text,
                    "parts": [{"type": "text", "text": text}],
                },
            ],
        }

        response = self.client.post(
            url="https://polymarket.com/api/grok/election-market-explanation",
            json=json_payload,
        )
        response.raise_for_status()

        parts = [p.strip() for p in response.text.split("**") if p.strip()]
        for i, part in enumerate(parts):
            if ":" in part and i != 0:
                print()
            print(part)

    def __enter__(self) -> object:
        return self

    def __exit__(self, exc_type: object, exc_val: object, exc_tb: object) -> None:
        self.client.close()



from __future__ import annotations

from datetime import datetime
from typing import Literal, Optional, cast

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    Json,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    field_validator,
)

from .common import EthAddress, FlexibleDatetime, Keccak256


class OptimizedImage(BaseModel):
    """Optimized image data."""

    model_config = ConfigDict(populate_by_name=True)

    id: Optional[str] = None
    image_url_source: Optional[str] = Field(None, alias="imageUrlSource")
    image_url_optimized: Optional[str] = Field(None, alias="imageUrlOptimized")
    image_size_kb_source: Optional[int] = Field(None, alias="imageSizeKbSource")
    image_size_kb_optimized: Optional[int] = Field(None, alias="imageSizeKbOptimized")
    image_optimized_complete: Optional[bool] = Field(
        None, alias="imageOptimizedComplete"
    )
    image_optimized_last_updated: Optional[str] = Field(
        None, alias="imageOptimizedLastUpdated"
    )
    rel_id: Optional[int] = Field(None, alias="relID")
    field: Optional[str] = Field(None, alias="field")
    relname: Optional[str] = None


class GammaMarket(BaseModel):
    """Market model."""

    model_config = ConfigDict(populate_by_name=True)

    id: Optional[str] = None
    condition_id: Optional[Keccak256] = Field(None, alias="conditionId")
    question_id: Optional[Keccak256] = Field(None, alias="questionID")
    slug: Optional[str] = Field(None, alias="slug")
    question: Optional[str] = Field(None, alias="question")
    twitter_card_image: Optional[str] = Field(None, alias="twitterCardImage")
    resolution_source: Optional[str] = Field(None, alias="resolutionSource")
    end_date: Optional[datetime] = Field(None, alias="endDate")
    category: Optional[str] = Field(None, alias="category")
    amm_type: Optional[str] = Field(None, alias="ammType")
    liquidity: Optional[float] = Field(None, alias="liquidity")
    sponsor_name: Optional[str] = Field(None, alias="sponsorName")
    sponsor_image: Optional[str] = Field(None, alias="sponsorImage")
    start_date: Optional[datetime] = Field(None, alias="startDate")
    x_axis_value: Optional[str] = Field(None, alias="xAxisValue")
    y_axis_value: Optional[str] = Field(None, alias="yAxisValue")
    denomination_token: Optional[str] = Field(None, alias="denominationToken")
    fee: Optional[str] = Field(None, alias="fee")
    fee_type: Optional[str] = Field(None, alias="feeType")
    image: Optional[str] = Field(None, alias="image")
    icon: Optional[str] = Field(None, alias="icon")
    lower_bound: Optional[str] = Field(None, alias="lowerBound")
    upper_bound: Optional[str] = Field(None, alias="upperBound")
    description: Optional[str] = Field(None, alias="description")
    outcomes: Optional[str] = Field(None, alias="outcomes")
    outcome_prices: Optional[Json[list[float]] | list[float]] = Field(
        None, alias="outcomePrices"
    )
    volume: Optional[str] = Field(None, alias="volume")
    active: Optional[bool] = Field(None, alias="active")
    market_type: Optional[str] = Field(None, alias="marketType")
    format_type: Optional[str] = Field(None, alias="formatType")
    lower_bound_date: Optional[datetime] = Field(None, alias="lowerBoundDate")
    upper_bound_date: Optional[datetime] = Field(None, alias="upperBoundDate")
    closed: Optional[bool] = Field(None, alias="closed")
    market_maker_address: Optional[str] = Field(None, alias="marketMakerAddress")
    created_by: Optional[int] = Field(None, alias="createdBy")
    updated_by: Optional[int] = Field(None, alias="updatedBy")
    created_at: Optional[datetime] = Field(None, alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")
    closed_time: Optional[str] = Field(None, alias="closedTime")
    wide_format: Optional[bool] = Field(None, alias="wideFormat")
    new: Optional[bool] = Field(None, alias="new")
    mailchimp_tag: Optional[str] = Field(None, alias="mailchimpTag")
    featured: Optional[bool] = Field(None, alias="featured")
    archived: Optional[bool] = Field(None, alias="archived")
    resolved_by: Optional[EthAddress] = Field(None, alias="resolvedBy")
    restricted: Optional[bool] = Field(None, alias="restricted")
    market_group: Optional[int] = Field(None, alias="marketGroup")
    group_item_title: Optional[str] = Field(None, alias="groupItemTitle")
    group_item_threshold: Optional[str] = Field(None, alias="groupItemThreshold")
    uma_end_date: Optional[FlexibleDatetime] = Field(None, alias="umaEndDate")
    enable_order_book: Optional[bool] = Field(None, alias="enableOrderBook")
    order_price_min_tick_size: Optional[float] = Field(
        None, alias="orderPriceMinTickSize"
    )
    order_min_size: Optional[float] = Field(None, alias="orderMinSize")
    uma_resolution_status: Optional[str] = Field(None, alias="umaResolutionStatus")
    curation_order: Optional[int] = Field(None, alias="curationOrder")
    volume_num: Optional[float] = Field(None, alias="volumeNum")
    liquidity_num: Optional[float] = Field(None, alias="liquidityNum")
    end_date_iso: Optional[datetime] = Field(None, alias="endDateIso")
    start_date_iso: Optional[datetime] = Field(None, alias="startDateIso")
    uma_end_date_iso: Optional[datetime] = Field(None, alias="umaEndDateIso")
    has_reviewed_dates: Optional[bool] = Field(None, alias="hasReviewedDates")
    ready_for_cron: Optional[bool] = Field(None, alias="readyForCron")
    comments_enabled: Optional[bool] = Field(None, alias="commentsEnabled")
    volume_24hr: Optional[float] = Field(None, alias="volume24hr")
    volume_1wk: Optional[float] = Field(None, alias="volume1wk")
    volume_1mo: Optional[float] = Field(None, alias="volume1mo")
    volume_1yr: Optional[float] = Field(None, alias="volume1yr")
    game_start_time: Optional[str] = Field(None, alias="gameStartTime")
    seconds_delay: Optional[int] = Field(None, alias="secondsDelay")
    token_ids: Optional[Json[list[str]] | list[str]] = Field(None, alias="clobTokenIds")
    disqus_thread: Optional[str] = Field(None, alias="disqusThread")
    short_outcomes: Optional[str] = Field(None, alias="shortOutcomes")
    team_a_id: Optional[str] = Field(None, alias="teamAID")
    team_b_id: Optional[str] = Field(None, alias="teamBID")
    uma_bond: Optional[str] = Field(None, alias="umaBond")
    uma_reward: Optional[str] = Field(None, alias="umaReward")
    fpmm_live: Optional[bool] = Field(None, alias="fpmmLive")
    volume_24hr_amm: Optional[float] = Field(None, alias="volume24hrAmm")
    volume_1wk_amm: Optional[float] = Field(None, alias="volume1wkAmm")
    volume_1mo_amm: Optional[float] = Field(None, alias="volume1moAmm")
    volume_1yr_amm: Optional[float] = Field(None, alias="volume1yrAmm")
    volume_24hr_clob: Optional[float] = Field(None, alias="volume24hrClob")
    volume_1wk_clob: Optional[float] = Field(None, alias="volume1wkClob")
    volume_1mo_clob: Optional[float] = Field(None, alias="volume1moClob")
    volume_1yr_clob: Optional[float] = Field(None, alias="volume1yrClob")
    volume_amm: Optional[float] = Field(None, alias="volumeAmm")
    volume_clob: Optional[float] = Field(None, alias="volumeClob")
    liquidity_amm: Optional[float] = Field(None, alias="liquidityAmm")
    liquidity_clob: Optional[float] = Field(None, alias="liquidityClob")
    maker_base_fee: Optional[int] = Field(None, alias="makerBaseFee")
    taker_base_fee: Optional[int] = Field(None, alias="takerBaseFee")
    custom_liveness: Optional[int] = Field(None, alias="customLiveness")
    accepting_orders: Optional[bool] = Field(None, alias="acceptingOrders")
    notifications_enabled: Optional[bool] = Field(None, alias="notificationsEnabled")
    requires_translation: Optional[bool] = Field(None, alias="requiresTranslation")
    score: Optional[int] = Field(None, alias="score")
    image_optimized: Optional[OptimizedImage] = Field(None, alias="imageOptimized")
    icon_optimized: Optional[OptimizedImage] = Field(None, alias="iconOptimized")
    events: Optional[list[Event]] = None
    categories: Optional[list[Category]] = Field(None, alias="categories")
    tags: Optional[list[Tag]] = Field(None, alias="tags")
    creator: Optional[str] = Field(None, alias="creator")
    ready: Optional[bool] = Field(None, alias="ready")
    funded: Optional[bool] = Field(None, alias="funded")
    past_slugs: Optional[str] = Field(None, alias="pastSlugs")
    ready_timestamp: Optional[datetime] = Field(None, alias="readyTimestamp")
    funded_timestamp: Optional[datetime] = Field(None, alias="fundedTimestamp")
    accepting_orders_timestamp: Optional[datetime] = Field(
        None, alias="acceptingOrdersTimestamp"
    )
    competitive: Optional[float] = Field(None, alias="competitive")
    rewards_min_size: Optional[float] = Field(None, alias="rewardsMinSize")
    rewards_max_spread: Optional[float] = Field(None, alias="rewardsMaxSpread")
    spread: Optional[float] = Field(None, alias="spread")
    automatically_resolved: Optional[bool] = Field(None, alias="automaticallyResolved")
    one_day_price_change: Optional[float] = Field(None, alias="oneDayPriceChange")
    one_hour_price_change: Optional[float] = Field(None, alias="oneHourPriceChange")
    one_week_price_change: Optional[float] = Field(None, alias="oneWeekPriceChange")
    one_month_price_change: Optional[float] = Field(None, alias="oneMonthPriceChange")
    one_year_price_change: Optional[float] = Field(None, alias="oneYearPriceChange")
    last_trade_price: Optional[float] = Field(None, alias="lastTradePrice")
    best_bid: Optional[float] = Field(None, alias="bestBid")
    best_ask: Optional[float] = Field(None, alias="bestAsk")
    automatically_active: Optional[bool] = Field(None, alias="automaticallyActive")
    clear_book_on_start: Optional[bool] = Field(None, alias="clearBookOnStart")
    chart_color: Optional[str] = Field(None, alias="chartColor")
    series_color: Optional[str] = Field(None, alias="seriesColor")
    show_gmp_series: Optional[bool] = Field(None, alias="showGmpSeries")
    show_gmp_outcome: Optional[bool] = Field(None, alias="showGmpOutcome")
    manual_activation: Optional[bool] = Field(None, alias="manualActivation")
    neg_risk_other: Optional[bool] = Field(None, alias="negRiskOther")
    game_id: Optional[str] = Field(None, alias="gameId")
    group_item_range: Optional[str] = Field(None, alias="groupItemRange")
    sports_market_type: Optional[str] = Field(None, alias="sportsMarketType")
    line: Optional[float] = Field(None, alias="line")
    uma_resolution_statuses: Optional[str] = Field(None, alias="umaResolutionStatuses")
    pending_deployment: Optional[bool] = Field(None, alias="pendingDeployment")
    deploying: Optional[bool] = Field(None, alias="deploying")
    deploying_timestamp: Optional[datetime] = Field(None, alias="deployingTimestamp")
    scheduled_deployment_timestamp: Optional[datetime] = Field(
        None, alias="scheduledDeploymentTimestamp"
    )
    rfq_enabled: Optional[bool] = Field(None, alias="rfqEnabled")
    event_start_time: Optional[datetime] = Field(None, alias="eventStartTime")
    clob_rewards: Optional[list[ClobReward]] = Field(None, alias="clobRewards")
    submitted_by: Optional[str] = None
    approved: Optional[bool] = None
    pager_duty_notification_enabled: Optional[bool] = Field(
        None, alias="pagerDutyNotificationEnabled"
    )
    holding_rewards_enabled: Optional[bool] = Field(None, alias="holdingRewardsEnabled")
    fees_enabled: Optional[bool] = Field(None, alias="feesEnabled")
    cyom: Optional[bool] = Field(None, alias="cyom")
    neg_risk: Optional[bool] = Field(None, alias="negRisk")
    neg_risk_request_id: Optional[str] = Field(None, alias="negRiskRequestID")
    neg_risk_market_id: Optional[str] = Field(None, alias="negRiskMarketID")
    sent_discord: Optional[bool] = Field(None, alias="sentDiscord")
    twitter_card_last_refreshed: Optional[datetime] = Field(
        None, alias="twitterCardLastRefreshed"
    )
    twitter_card_last_validated: Optional[datetime] = Field(
        None, alias="twitterCardLastValidated"
    )
    twitter_card_location: Optional[str] = Field(None, alias="twitterCardLocation")

    @field_validator("condition_id", mode="wrap")
    @classmethod
    def validate_condition_id(
        cls,
        value: str,
        handler: ValidatorFunctionWrapHandler,
        info: ValidationInfo,
    ) -> str:
        try:
            # First attempt standard Keccak256 validation
            return cast("str", handler(value))
        except ValueError:
            active = info.data.get("active", False)

            # Only allow empty string when inactive
            if not active and value == "":
                return value

            # Re-raise original error for other cases
            raise


class Series(BaseModel):
    """Series model."""

    model_config = ConfigDict(populate_by_name=True)

    id: Optional[str] = Field(None, alias="id")
    ticker: Optional[str] = Field(None, alias="ticker")
    slug: Optional[str] = Field(None, alias="slug")
    title: Optional[str] = Field(None, alias="title")
    subtitle: Optional[str] = Field(None, alias="subtitle")
    series_type: Optional[str] = Field(None, alias="seriesType")
    recurrence: Optional[str] = Field(None, alias="recurrence")
    description: Optional[str] = Field(None, alias="description")
    image: Optional[str] = Field(None, alias="image")
    icon: Optional[str] = Field(None, alias="icon")
    layout: Optional[str] = Field(None, alias="layout")
    active: Optional[bool] = Field(None, alias="active")
    closed: Optional[bool] = Field(None, alias="closed")
    archived: Optional[bool] = Field(None, alias="archived")
    new: Optional[bool] = Field(None, alias="new")
    featured: Optional[bool] = Field(None, alias="featured")
    restricted: Optional[bool] = Field(None, alias="restricted")
    is_template: Optional[bool] = Field(None, alias="isTemplate")
    template_variables: Optional[bool] = Field(None, alias="templateVariables")
    published_at: Optional[FlexibleDatetime] = Field(None, alias="publishedAt")
    created_by: Optional[str] = Field(None, alias="createdBy")
    updated_by: Optional[str] = Field(None, alias="updatedBy")
    created_at: Optional[datetime] = Field(None, alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")
    comments_enabled: Optional[bool] = Field(None, alias="commentsEnabled")
    competitive: Optional[str] = Field(None, alias="competitive")
    volume_24hr: Optional[float] = Field(None, alias="volume24hr")
    volume: Optional[float] = Field(None, alias="volume")
    liquidity: Optional[float] = Field(None, alias="liquidity")
    start_date: Optional[datetime] = Field(None, alias="startDate")
    pyth_token_id: Optional[str] = Field(None, alias="pythTokenID")
    cg_asset_name: Optional[str] = Field(None, alias="cgAssetName")
    score: Optional[int] = Field(None, alias="score")
    requires_translation: Optional[bool] = Field(None, alias="requiresTranslation")
    events: Optional[list[Event]] = Field(None, alias="events")
    collections: Optional[list[Collection]] = Field(None, alias="collections")
    categories: Optional[list[Category]] = Field(None, alias="categories")
    tags: Optional[list[Tag]] = Field(None, alias="tags")
    comment_count: Optional[int] = Field(None, alias="commentCount")
    chats: Optional[list[Chat]] = Field(None, alias="chats")


class Category(BaseModel):
    """Category model."""

    model_config = ConfigDict(populate_by_name=True)

    id: Optional[str] = Field(None, alias="id")
    label: Optional[str] = Field(None, alias="label")
    parent_category: Optional[str] = Field(None, alias="parentCategory")
    slug: Optional[str] = Field(None, alias="slug")
    published_at: Optional[str] = Field(None, alias="publishedAt")
    created_by: Optional[str] = Field(None, alias="createdBy")
    updated_by: Optional[str] = Field(None, alias="updatedBy")
    created_at: Optional[datetime] = Field(None, alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")


class Tag(BaseModel):
    """Tag model."""

    model_config = ConfigDict(populate_by_name=True)

    id: Optional[str] = Field(None, alias="id")
    label: Optional[str] = Field(None, alias="label")
    slug: Optional[str] = Field(None, alias="slug")
    force_show: Optional[bool] = Field(None, alias="forceShow")
    published_at: Optional[str] = Field(None, alias="publishedAt")
    created_by: Optional[int] = Field(None, alias="createdBy")
    updated_by: Optional[int] = Field(None, alias="updatedBy")
    created_at: Optional[datetime] = Field(None, alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")
    force_hide: Optional[bool] = Field(None, alias="forceHide")
    is_carousel: Optional[bool] = Field(None, alias="isCarousel")
    requires_translation: Optional[bool] = Field(None, alias="requiresTranslation")


class TagRelation(BaseModel):
    """Tag relation model."""

    model_config = ConfigDict(populate_by_name=True)

    id: int = Field(alias="id")
    tag_id: int = Field(alias="tagID")
    related_tag_id: int = Field(alias="relatedTagID")
    rank: int


class Chat(BaseModel):
    """Chat model."""

    model_config = ConfigDict(populate_by_name=True)

    id: Optional[str] = Field(None, alias="id")
    channel_id: Optional[str] = Field(None, alias="channelId")
    channel_name: Optional[str] = Field(None, alias="channelName")
    channel_image: Optional[str] = Field(None, alias="channelImage")
    live: Optional[bool] = Field(None, alias="live")
    start_time: Optional[datetime] = Field(None, alias="startTime")
    end_time: Optional[datetime] = Field(None, alias="endTime")


class Collection(BaseModel):
    """Collection model."""

    model_config = ConfigDict(populate_by_name=True)

    id: Optional[str] = Field(None, alias="id")
    ticker: Optional[str] = Field(None, alias="ticker")
    slug: Optional[str] = Field(None, alias="slug")
    title: Optional[str] = Field(None, alias="title")
    subtitle: Optional[str] = Field(None, alias="subtitle")
    collection_type: Optional[str] = Field(None, alias="collectionType")
    description: Optional[str] = Field(None, alias="description")
    tags: Optional[str] = Field(None, alias="tags")
    image: Optional[str] = Field(None, alias="image")
    icon: Optional[str] = Field(None, alias="icon")
    header_image: Optional[str] = Field(None, alias="headerImage")
    layout: Optional[str] = Field(None, alias="layout")
    active: Optional[bool] = Field(None, alias="active")
    closed: Optional[bool] = Field(None, alias="closed")
    archived: Optional[bool] = Field(None, alias="archived")
    new: Optional[bool] = Field(None, alias="new")
    featured: Optional[bool] = Field(None, alias="featured")
    restricted: Optional[bool] = Field(None, alias="restricted")
    is_template: Optional[bool] = Field(None, alias="isTemplate")
    template_variables: Optional[str] = Field(None, alias="templateVariables")
    published_at: Optional[str] = Field(None, alias="publishedAt")
    created_by: Optional[str] = Field(None, alias="createdBy")
    updated_by: Optional[str] = Field(None, alias="updatedBy")
    created_at: Optional[datetime] = Field(None, alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")
    comments_enabled: Optional[bool] = Field(None, alias="commentsEnabled")
    image_optimized: Optional[OptimizedImage] = Field(None, alias="imageOptimized")
    icon_optimized: Optional[OptimizedImage] = Field(None, alias="iconOptimized")
    header_image_optimized: Optional[OptimizedImage] = Field(
        None, alias="headerImageOptimized"
    )


class Creator(BaseModel):
    """Event Creator model."""

    model_config = ConfigDict(populate_by_name=True)

    id: Optional[str] = Field(None, alias="id")
    creator_name: Optional[str] = Field(None, alias="creatorName")
    creator_handle: Optional[str] = Field(None, alias="creatorHandle")
    creator_url: Optional[str] = Field(None, alias="creatorUrl")
    creator_image: Optional[str] = Field(None, alias="creatorImage")
    created_at: Optional[datetime] = Field(None, alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")


class Template(BaseModel):
    """Template model."""

    model_config = ConfigDict(populate_by_name=True)

    id: Optional[str] = Field(None, alias="id")
    event_title: Optional[str] = Field(None, alias="eventTitle")
    event_slug: Optional[str] = Field(None, alias="eventSlug")
    event_image: Optional[str] = Field(None, alias="eventImage")
    market_title: Optional[str] = Field(None, alias="marketTitle")
    description: Optional[str] = Field(None, alias="description")
    resolution_source: Optional[str] = Field(None, alias="resolutionSource")
    neg_risk: Optional[bool] = Field(None, alias="negRisk")
    sort_by: Optional[str] = Field(None, alias="sortBy")
    show_market_images: Optional[bool] = Field(None, alias="showMarketImages")
    series_slug: Optional[str] = Field(None, alias="seriesSlug")
    outcomes: Optional[str] = Field(None, alias="outcomes")


class ClobReward(BaseModel):
    """Reward model."""

    model_config = ConfigDict(populate_by_name=True)

    id: int
    condition_id: Keccak256 = Field(alias="conditionId")
    asset_address: str = Field(alias="assetAddress")
    rewards_amount: float = Field(alias="rewardsAmount")
    rewards_daily_rate: Optional[float] = Field(None, alias="rewardsDailyRate")
    start_date: datetime = Field(alias="startDate")
    end_date: datetime = Field(alias="endDate")


class Team(BaseModel):
    """Team model."""

    model_config = ConfigDict(populate_by_name=True)

    id: int
    name: str
    league: str
    record: Optional[str] = None
    logo: str
    abbreviation: str
    alias: Optional[str] = None
    created_at: datetime = Field(alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")


class Sport(BaseModel):
    """Sport model."""

    model_config = ConfigDict(populate_by_name=True)

    sport: str
    image: Optional[str] = None
    resolution: Optional[str] = None
    ordering: Optional[Literal["home", "away"]] = None
    tags: Optional[list[int]] = None
    series: Optional[int] = None

    @field_validator("tags", mode="before")
    @classmethod
    def split_string_to_int_list(cls, v: str | list[int]) -> list[int]:
        if isinstance(v, str):
            return [int(i) for i in v.split(",")]
        return v


class Event(BaseModel):
    """Event model."""

    model_config = ConfigDict(populate_by_name=True)

    id: int = Field(alias="id")
    ticker: Optional[str] = Field(None, alias="ticker")
    slug: Optional[str] = Field(None, alias="slug")
    title: Optional[str] = Field(None, alias="title")
    subtitle: Optional[str] = Field(None, alias="subtitle")
    description: Optional[str] = Field(None, alias="description")
    resolution_source: Optional[str] = Field(None, alias="resolutionSource")
    start_date: Optional[datetime] = Field(None, alias="startDate")
    creation_date: Optional[datetime] = Field(None, alias="creationDate")
    end_date: Optional[datetime] = Field(None, alias="endDate")
    image: Optional[str] = Field(None, alias="image")
    icon: Optional[str] = Field(None, alias="icon")
    active: Optional[bool] = Field(None, alias="active")
    closed: Optional[bool] = Field(None, alias="closed")
    archived: Optional[bool] = Field(None, alias="archived")
    new: Optional[bool] = Field(None, alias="new")
    featured: Optional[bool] = Field(None, alias="featured")
    restricted: Optional[bool] = Field(None, alias="restricted")
    requires_translation: Optional[bool] = Field(None, alias="requiresTranslation")
    liquidity: Optional[float] = Field(None, alias="liquidity")
    volume: Optional[float] = Field(None, alias="volume")
    open_interest: Optional[int] = Field(None, alias="openInterest")
    sort_by: Optional[str] = Field(None, alias="sortBy")
    category: Optional[str] = Field(None, alias="category")
    subcategory: Optional[str] = Field(None, alias="subcategory")
    is_template: Optional[bool] = Field(None, alias="isTemplate")
    template_variables: Optional[str] = Field(None, alias="templateVariables")
    published_at: Optional[FlexibleDatetime] = Field(None, alias="publishedAt")
    created_by: Optional[str] = Field(None, alias="createdBy")
    updated_by: Optional[str] = Field(None, alias="updatedBy")
    created_at: Optional[datetime] = Field(None, alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")
    comments_enabled: Optional[bool] = Field(None, alias="commentsEnabled")
    competitive: Optional[float] = Field(None, alias="competitive")
    volume_24hr: Optional[float] = Field(None, alias="volume24hr")
    volume_1wk: Optional[float] = Field(None, alias="volume1wk")
    volume_1mo: Optional[float] = Field(None, alias="volume1mo")
    volume_1yr: Optional[float] = Field(None, alias="volume1yr")
    featured_image: Optional[str] = Field(None, alias="featuredImage")
    disqus_thread: Optional[str] = Field(None, alias="disqusThread")
    parent_event: Optional[str] = Field(None, alias="parentEvent")
    enable_order_book: Optional[bool] = Field(None, alias="enableOrderBook")
    liquidity_amm: Optional[float] = Field(None, alias="liquidityAmm")
    liquidity_clob: Optional[float] = Field(None, alias="liquidityClob")
    neg_risk: Optional[bool] = Field(None, alias="negRisk")
    neg_risk_augmented: Optional[bool] = Field(None, alias="negRiskAugmented")
    neg_risk_market_id: Optional[str] = Field(None, alias="negRiskMarketID")
    neg_risk_fee_bips: Optional[int] = Field(None, alias="negRiskFeeBips")
    comment_count: Optional[int] = Field(None, alias="commentCount")
    image_optimized: Optional[OptimizedImage] = Field(None, alias="imageOptimized")
    icon_optimized: Optional[OptimizedImage] = Field(None, alias="iconOptimized")
    featured_image_optimized: Optional[OptimizedImage] = Field(
        None, alias="featuredImageOptimized"
    )
    sub_events: Optional[list[str]] = Field(None, alias="subEvents")
    markets: Optional[list[GammaMarket]] = Field(None, alias="markets")
    series: Optional[list[Series]] = Field(None, alias="series")
    categories: Optional[list[Category]] = Field(None, alias="categories")
    collections: Optional[list[Collection]] = Field(None, alias="collections")
    tags: Optional[list[Tag]] = Field(None, alias="tags")
    cyom: Optional[bool] = Field(None, alias="cyom")
    closed_time: Optional[datetime] = Field(None, alias="closedTime")
    show_all_outcomes: Optional[bool] = Field(None, alias="showAllOutcomes")
    show_market_images: Optional[bool] = Field(None, alias="showMarketImages")
    automatically_resolved: Optional[bool] = Field(None, alias="automaticallyResolved")
    enable_neg_risk: Optional[bool] = Field(None, alias="enableNegRisk")
    automatically_active: Optional[bool] = Field(None, alias="automaticallyActive")
    event_date: Optional[datetime] = Field(None, alias="eventDate")
    start_time: Optional[datetime] = Field(None, alias="startTime")
    event_week: Optional[int] = Field(None, alias="eventWeek")
    series_slug: Optional[str] = Field(None, alias="seriesSlug")
    score: Optional[str] = Field(None, alias="score")
    elapsed: Optional[str] = Field(None, alias="elapsed")
    period: Optional[str] = Field(None, alias="period")
    live: Optional[bool] = Field(None, alias="live")
    ended: Optional[bool] = Field(None, alias="ended")
    finished_timestamp: Optional[datetime] = Field(None, alias="finishedTimestamp")
    gmp_chart_mode: Optional[str] = Field(None, alias="gmpChartMode")
    event_creators: Optional[list[Creator]] = Field(None, alias="eventCreators")
    tweet_count: Optional[int] = Field(None, alias="tweetCount")
    chats: Optional[list[Chat]] = Field(None, alias="chats")
    featured_order: Optional[int] = Field(None, alias="featuredOrder")
    estimate_value: Optional[bool] = Field(None, alias="estimateValue")
    cant_estimate: Optional[bool] = Field(None, alias="cantEstimate")
    estimated_value: Optional[str] = Field(None, alias="estimatedValue")
    templates: Optional[list[Template]] = Field(None, alias="templates")
    spreads_main_line: Optional[float] = Field(None, alias="spreadsMainLine")
    totals_main_line: Optional[float] = Field(None, alias="totalsMainLine")
    carousel_map: Optional[str] = Field(None, alias="carouselMap")
    pending_deployment: Optional[bool] = Field(None, alias="pendingDeployment")
    deploying: Optional[bool] = Field(None, alias="deploying")
    deploying_timestamp: Optional[datetime] = Field(None, alias="deployingTimestamp")
    scheduled_deployment_timestamp: Optional[datetime] = Field(
        None, alias="scheduledDeploymentTimestamp"
    )
    game_status: Optional[str] = Field(None, alias="gameStatus")
    cumulative_markets: Optional[int] = Field(None, alias="cumulativeMarkets")
    event_metadata: Optional[dict[str, object]] = Field(None, alias="eventMetadata")


class ProfilePosition(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    token_id: int = Field(alias="tokenId")
    size: Optional[float] = Field(None, alias="positionSize")

    @field_validator("size", mode="before")
    @classmethod
    def normalize_size(cls, v: str | float) -> float:
        if isinstance(v, str):
            return int(v) / 10**6
        return v


class Profile(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    name: Optional[str] = None
    pseudonym: Optional[str] = None
    display_username_public: Optional[bool] = Field(None, alias="displayUsernamePublic")
    proxy_wallet: Optional[EthAddress] = Field(None, alias="proxyWallet")
    base_address: Optional[EthAddress] = Field(None, alias="baseAddress")
    profile_image: Optional[str] = Field(None, alias="profileImage")
    positions: Optional[list[ProfilePosition]] = Field(None, alias="positions")


class Comment(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    id: str
    body: str
    parent_entity_type: Literal["Event", "Series", "market"] = Field(
        alias="parentEntityType"
    )
    parent_entity_id: int = Field(alias="parentEntityID")
    parent_comment_id: Optional[str] = Field(None, alias="parentCommentID")
    user_address: str = Field(alias="userAddress")
    reply_address: Optional[str] = Field(None, alias="replyAddress")
    created_at: datetime = Field(alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")
    profile: Optional[Profile] = None
    reactions: Optional[list[Reaction]] = None
    report_count: Optional[int] = Field(None, alias="reportCount")
    reaction_count: Optional[int] = Field(None, alias="reactionCount")


class Reaction(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    id: str
    comment_id: int = Field(alias="commentID")
    reaction_type: str = Field(alias="reactionType")
    icon: Optional[str] = None
    user_address: str = Field(alias="userAddress")
    created_at: datetime = Field(alias="createdAt")
    profile: Optional[Profile] = None


class Pagination(BaseModel):
    has_more: bool = Field(alias="hasMore")
    total_results: int = Field(alias="totalResults")


class SearchResult(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    events: Optional[list[Event]] = None
    tags: Optional[list[Tag]] = None
    profiles: Optional[list[Profile]] = None
    pagination: Pagination


Event.model_rebuild()
Series.model_rebuild()
GammaMarket.model_rebuild()
SearchResult.model_rebuild()



from typing import TYPE_CHECKING

from ens.ens import ChecksumAddress
from eth_account import Account
from eth_typing import HexStr

if TYPE_CHECKING:
    from eth_account.datastructures import SignedMessage
    from eth_account.signers.local import LocalAccount


class Signer:
    def __init__(self, private_key: str, chain_id: int):
        if private_key is None:
            msg = "private_key must not be None"
            raise ValueError(msg)
        if chain_id is None:
            msg = "chain_id must not be None"
            raise ValueError(msg)

        self.private_key = private_key
        self.account: LocalAccount = Account.from_key(private_key)
        self.chain_id = chain_id

    def address(self) -> ChecksumAddress:
        return self.account.address

    def get_chain_id(self) -> int:
        return self.chain_id

    def sign(self, message_hash: HexStr | bytes | int) -> str:
        """Signs a message hash."""
        signed_hash: SignedMessage = Account.unsafe_sign_hash(
            message_hash, self.private_key
        )
        return signed_hash.signature.hex()



import re
from datetime import UTC, datetime
from typing import Annotated, Any, cast

from dateutil import parser
from eth_typing import ChecksumAddress, HexStr
from hexbytes import HexBytes
from pydantic import AfterValidator, BaseModel, BeforeValidator, ConfigDict, Field


def parse_flexible_datetime(v: str | datetime) -> datetime:
    """Parse datetime from multiple formats using dateutil."""
    if v in {"NOW*()", "NOW()"}:
        return datetime.fromtimestamp(0, tz=UTC)

    if isinstance(v, str):
        parsed = parser.parse(v)
        if not isinstance(parsed, datetime):
            msg = f"Failed to parse '{v}' as datetime, got {type(parsed)}"
            raise TypeError(msg)
        return parsed
    return v


# Validate format: 0x followed by 64 hex characters
KECCAK256_REGEX = re.compile(r"^0x[a-fA-F0-9]{64}$")


def validate_keccak256(v: str | HexBytes | bytes) -> str:
    """Validate and normalize Keccak256 hash format."""
    # Convert HexBytes/bytes to string
    if isinstance(v, HexBytes | bytes):
        v = v.hex()

    if not v.startswith("0x"):
        v = "0x" + v

    if not KECCAK256_REGEX.match(v):
        msg = f"Invalid Keccak256 hash format: {v}"
        raise ValueError(msg)

    return v


CHECKSUM_ADDRESS_REGEX = re.compile(r"^0x[a-fA-F0-9]{40}$")


def validate_eth_address(v: str | HexBytes | bytes) -> ChecksumAddress:
    """Validate and normalize Ethereum address format."""
    # Convert HexBytes/bytes to string
    if isinstance(v, HexBytes | bytes):
        v = v.hex()

    if not v.startswith("0x"):
        v = "0x" + v

    # Validate format: 0x followed by 40 hex characters
    if not CHECKSUM_ADDRESS_REGEX.match(v):
        msg = f"Invalid Ethereum address format: {v}"
        raise ValueError(msg)

    return cast("ChecksumAddress", v)


def hexbytes_to_str(v: HexBytes | bytes | str) -> HexStr:
    """Convert HexBytes to hex string with 0x prefix."""
    if isinstance(v, HexBytes):
        hex_str = v.hex()
        return cast("HexStr", hex_str if hex_str.startswith("0x") else f"0x{hex_str}")
    if isinstance(v, bytes):
        return cast("HexStr", "0x" + v.hex())
    if isinstance(v, str) and not v.startswith("0x"):
        return cast("HexStr", f"0x{v}")
    return cast("HexStr", v)


def validate_keccak_or_padded(v: Any) -> HexStr:
    """
    Validate Keccak256 or accept padded addresses (32 bytes with leading zeros).

    Some log topics are padded addresses, not proper Keccak256 hashes.
    """
    # First convert HexBytes/bytes to string with 0x prefix
    if isinstance(v, HexBytes | bytes):
        v = v.hex()

    # Ensure it's a string
    if not isinstance(v, str):
        msg = f"Expected string or bytes, got {type(v)}"
        raise TypeError(msg)

    # Add 0x prefix if missing
    v_str: HexStr = cast("HexStr", "0x" + v if not v.startswith("0x") else v)

    if KECCAK256_REGEX.match(v_str):
        return v_str

    msg = f"Invalid hash format: expected 66 characters (0x + 64 hex), got {len(v_str)}: {v_str}"
    raise ValueError(msg)


FlexibleDatetime = Annotated[datetime, BeforeValidator(parse_flexible_datetime)]
EthAddress = Annotated[ChecksumAddress, AfterValidator(validate_eth_address)]
Keccak256 = Annotated[str, AfterValidator(validate_keccak256)]
HexString = Annotated[str, BeforeValidator(hexbytes_to_str)]
Keccak256OrPadded = Annotated[str, BeforeValidator(validate_keccak_or_padded)]
EmptyString = Annotated[str, Field(pattern=r"^$", description="An empty string")]


class TimeseriesPoint(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    value: float = Field(alias="p")
    timestamp: datetime = Field(alias="t")



"""
Type definitions for Polymarket APIs.

This module contains all the Pydantic models and type definitions used across
the Polymarket APIs.
"""

from __future__ import annotations

from importlib import import_module
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .clob_types import (
        ApiCreds,
        AssetType,
        BidAsk,
        BookParams,
        ClobMarket,
        ContractConfig,
        CreateOrderOptions,
        DailyEarnedReward,
        MarketOrderArgs,
        MarketRewards,
        Midpoint,
        OpenOrder,
        OrderArgs,
        OrderBookSummary,
        OrderCancelResponse,
        OrderPostResponse,
        OrderType,
        PaginatedResponse,
        PartialCreateOrderOptions,
        PolygonTrade,
        PostOrdersArgs,
        Price,
        PriceHistory,
        RewardMarket,
        Spread,
        TickSize,
        Token,
        TokenBidAsk,
        TokenBidAskDict,
        TokenValue,
        TokenValueDict,
    )
    from .common import (
        EmptyString,
        EthAddress,
        FlexibleDatetime,
        Keccak256,
        TimeseriesPoint,
    )
    from .data_types import (
        Activity,
        ActivityType,
        Holder,
        HolderResponse,
        Position,
        Trade,
        User,
        UserMetric,
        UserRank,
        ValueResponse,
    )
    from .gamma_types import ClobReward, Event, GammaMarket, Pagination, Series, Tag
    from .websockets_types import (
        ActivityOrderMatchEvent,
        ActivityTradeEvent,
        AssetPriceSubscribeEvent,
        AssetPriceUpdateEvent,
        BestBidAskEvent,
        CommentEvent,
        ErrorEvent,
        LastTradePriceEvent,
        LiveDataEvents,
        MarketEvents,
        MarketResolvedEvent,
        NewMarketEvent,
        OrderBookSummaryEvent,
        OrderEvent,
        PriceChangeEvent,
        QuoteEvent,
        ReactionEvent,
        RequestEvent,
        SportsGameUpdate,
        TickSizeChangeEvent,
        TradeEvent,
        UserEvents,
    )

__all__ = [
    "Activity",
    "ActivityOrderMatchEvent",
    "ActivityTradeEvent",
    "ActivityType",
    "ApiCreds",
    "AssetPriceSubscribeEvent",
    "AssetPriceUpdateEvent",
    "AssetType",
    "BestBidAskEvent",
    "BidAsk",
    "BookParams",
    "ClobMarket",
    "ClobReward",
    "CommentEvent",
    "ContractConfig",
    "CreateOrderOptions",
    "DailyEarnedReward",
    "EmptyString",
    "ErrorEvent",
    "EthAddress",
    "Event",
    "FlexibleDatetime",
    "GammaMarket",
    "Holder",
    "HolderResponse",
    "Keccak256",
    "LastTradePriceEvent",
    "LiveDataEvents",
    "MarketEvents",
    "MarketOrderArgs",
    "MarketResolvedEvent",
    "MarketRewards",
    "Midpoint",
    "NewMarketEvent",
    "OpenOrder",
    "OrderArgs",
    "OrderBookSummary",
    "OrderBookSummaryEvent",
    "OrderCancelResponse",
    "OrderEvent",
    "OrderPostResponse",
    "OrderType",
    "PaginatedResponse",
    "Pagination",
    "PartialCreateOrderOptions",
    "PolygonTrade",
    "Position",
    "PostOrdersArgs",
    "Price",
    "PriceChangeEvent",
    "PriceHistory",
    "QuoteEvent",
    "ReactionEvent",
    "RequestEvent",
    "RewardMarket",
    "Series",
    "SportsGameUpdate",
    "Spread",
    "Tag",
    "TickSize",
    "TickSizeChangeEvent",
    "TimeseriesPoint",
    "Token",
    "TokenBidAsk",
    "TokenBidAskDict",
    "TokenValue",
    "TokenValueDict",
    "Trade",
    "TradeEvent",
    "User",
    "UserEvents",
    "UserMetric",
    "UserRank",
    "ValueResponse",
]

_EXPORT_MAP = {
    "Activity": ".data_types",
    "ActivityOrderMatchEvent": ".websockets_types",
    "ActivityTradeEvent": ".websockets_types",
    "ActivityType": ".data_types",
    "ApiCreds": ".clob_types",
    "AssetPriceSubscribeEvent": ".websockets_types",
    "AssetPriceUpdateEvent": ".websockets_types",
    "AssetType": ".clob_types",
    "BestBidAskEvent": ".websockets_types",
    "BidAsk": ".clob_types",
    "BookParams": ".clob_types",
    "ClobMarket": ".clob_types",
    "ClobReward": ".gamma_types",
    "CommentEvent": ".websockets_types",
    "ContractConfig": ".clob_types",
    "CreateOrderOptions": ".clob_types",
    "DailyEarnedReward": ".clob_types",
    "EmptyString": ".common",
    "ErrorEvent": ".websockets_types",
    "EthAddress": ".common",
    "Event": ".gamma_types",
    "FlexibleDatetime": ".common",
    "GammaMarket": ".gamma_types",
    "Holder": ".data_types",
    "HolderResponse": ".data_types",
    "Keccak256": ".common",
    "LastTradePriceEvent": ".websockets_types",
    "LiveDataEvents": ".websockets_types",
    "MarketEvents": ".websockets_types",
    "MarketOrderArgs": ".clob_types",
    "MarketResolvedEvent": ".websockets_types",
    "MarketRewards": ".clob_types",
    "Midpoint": ".clob_types",
    "NewMarketEvent": ".websockets_types",
    "OpenOrder": ".clob_types",
    "OrderArgs": ".clob_types",
    "OrderBookSummary": ".clob_types",
    "OrderBookSummaryEvent": ".websockets_types",
    "OrderCancelResponse": ".clob_types",
    "OrderEvent": ".websockets_types",
    "OrderPostResponse": ".clob_types",
    "OrderType": ".clob_types",
    "PaginatedResponse": ".clob_types",
    "Pagination": ".gamma_types",
    "PartialCreateOrderOptions": ".clob_types",
    "PolygonTrade": ".clob_types",
    "Position": ".data_types",
    "PostOrdersArgs": ".clob_types",
    "Price": ".clob_types",
    "PriceChangeEvent": ".websockets_types",
    "PriceHistory": ".clob_types",
    "QuoteEvent": ".websockets_types",
    "ReactionEvent": ".websockets_types",
    "RequestEvent": ".websockets_types",
    "RewardMarket": ".clob_types",
    "Series": ".gamma_types",
    "SportsGameUpdate": ".websockets_types",
    "Spread": ".clob_types",
    "Tag": ".gamma_types",
    "TickSize": ".clob_types",
    "TickSizeChangeEvent": ".websockets_types",
    "TimeseriesPoint": ".common",
    "Token": ".clob_types",
    "TokenBidAsk": ".clob_types",
    "TokenBidAskDict": ".clob_types",
    "TokenValue": ".clob_types",
    "TokenValueDict": ".clob_types",
    "Trade": ".data_types",
    "TradeEvent": ".websockets_types",
    "User": ".data_types",
    "UserEvents": ".websockets_types",
    "UserMetric": ".data_types",
    "UserRank": ".data_types",
    "ValueResponse": ".data_types",
}


def __getattr__(name: str) -> Any:
    module_name = _EXPORT_MAP.get(name)
    if module_name is None:
        msg = f"module {__name__!r} has no attribute {name!r}"
        raise AttributeError(msg)

    module = import_module(module_name, __name__)
    value = getattr(module, name)
    globals()[name] = value
    return value
